package conv.convi;

import 	 jam.vector.*;
import 	 jam.operator.*;
import   rsf.vector.*;
import 	 sep.operator.SepBiLinearOperator;

/**
 * Internal convolution (as a bilinear operator)
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * <BR><BR>The map is   A(i,f) |---> o
 * <BR> where: <UL>
 * <LI> A = bilinear internal convolution operator
 * <LI> i = input
 * <LI> f = filter
 * <LI> o = output
 * </UL>
 * <BR><BR>
 * 
 * @author    Joel Schroeder
 */
public class InternalConvolution extends SepBiLinearOperator {
  public static final int kInput  = SepBiLinearOperator.kInput1;
  public static final int kFilter = SepBiLinearOperator.kInput2;
  public static final int kOutput = SepBiLinearOperator.kOutput;

  public InternalConvolution(int spc1, RsfSpace space1,
			     int spc2, RsfSpace space2) {
    super(null, null, null);

    if ((spc1 != spc2)
	&& setSpace(spc1, space1)
	&& setSpace(spc2, space2)) {
      if      (getRange()   == null)
	setRange(  createOutputSpace((RsfSpace) getDomain1(),
				     (RsfSpace) getDomain2()));
      else if (getDomain1() == null)
	setDomain1(createInputSpace ((RsfSpace) getDomain2(),
				     (RsfSpace) getRange()));
      else
	setDomain2(createFilterSpace((RsfSpace) getDomain1(),
				     (RsfSpace) getRange()));
      return;
    }

    System.err.println("Error: integers to InternalConvolution" + 
		       " constructor must be two of:"           + "\n" +
		       " kInput, kFilter, kOutput." + "\n"); 
  }

  /** copy constructor */
  protected InternalConvolution(InternalConvolution src) {
    super(src);
  }

  /** The domain and range spaces are shallow copies */
  public jam.JamObject copy() { return new InternalConvolution(this); }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (!(obj2 instanceof InternalConvolution)) return false;
    return super.equals((InternalConvolution) obj2);
  }

  public Operator getOp1(Vector x1) { return getIcaf((Rsf) x1); }
  public Operator getOp2(Vector x2) { return getIcai((Rsf) x2); }

  public Icaf     getIcaf(Rsf input) {
    if (!(getDomain1().hasMember(input)))
      System.err.println("Error in InternalConvolution.getIcaf(): " +
			 "Vector is not in domain1\n");
    return new IcafFactory(input).getIcafFromOutputSpace((RsfSpace) range);
  }

  public Icai     getIcai(Rsf filter) {
    if (!(getDomain2().hasMember(filter)))
      System.err.println("Error in InternalConvolution.getIcai(): " +
			 "Vector is not in domain2\n");
    return new IcaiFactory(filter).getIcaiFromOutputSpace((RsfSpace) range);
  }

  /**
   * <STRONG>WARNING: not yet implemented</STRONG>
   */
  public hasAdjoint getPartialAdjoint () {
    System.err.println("Error: InternalConvolution" +
			".getPartialAdjoint() not yet implemented.\n");
    return null; }

  // I/O
  public String toString() {
    return ("Internal Convolution:\n" + super.toString());
  }

  protected void apply(boolean add, Vector x, Vector f,
		       Vector y, int result) {
    staticApply(add, x, f, y, result);
  }

  static void staticApply(boolean add, Vector x, Vector f,
			  Vector y, int result) {
    int dim = ((RsfSpace) x.getSpace()).getDimension();

    switch (result) {
    case kInput:
      if (!add) x.zero();
      break;
    case kFilter:
      if (!add) f.zero();
      break;
    case kOutput:
      if (!add) y.zero();
      break;
    default:
      System.err.println("Error: integer argument \"result\" is invalid:\n" +
			 "       result = " + result);
      break;
    }

    switch (dim) {
    case   1:
      apply1D((Rsf1) x, (Rsf1) f, (Rsf1) y, result);
      break;
    case   2:
      apply2D((Rsf2) x, (Rsf2) f, (Rsf2) y, result);
      break;
    case 3:
      apply3D((Rsf3) x, (Rsf3) f, (Rsf3) y, result);
      break;
    case 4:
      apply4D((Rsf4) x, (Rsf4) f, (Rsf4) y, result);
      break;
    case 5:
      apply5D((Rsf5) x, (Rsf5) f, (Rsf5) y, result);
      break;
    case 6:
      apply6D((Rsf6) x, (Rsf6) f, (Rsf6) y, result);
      break;
    default:
      System.err.println("Error: Internal Convolution only supported" +
		         " up to 6 dimensions");
      break;
    }
  }

  static void apply1D(Rsf1 x, Rsf1 f, Rsf1 y, int result) {
    int fn = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn = ((RsfSpace) y.getSpace()).getNSamples(0);
    int x0, f0, y0;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for   (f0 = 0; f0 < fn; f0++)
      for (y0 = 0; y0 < yn; y0++) {
	x0 = y0 - f0 + fn - 1;
	switch (result) {
	case (kOutput):
	  yy[y0] += ff[f0] * xx[x0];
	  break;
	case (kInput ):
	  xx[x0] += yy[y0] * ff[f0];
	  break;
	case (kFilter):
	  ff[f0] += xx[x0] * yy[y0];
	  break;
	}
      }
  }

  static void apply2D(Rsf2 x, Rsf2 f, Rsf2 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int x0, x1;
    float[][] yy = y.d, ff = f.d, xx = x.d;
    for     (int f1 = 0; f1 < fn1; f1++)
      for   (int y1 = 0; y1 < yn1; y1++) { x1 = y1 - f1 + fn1 - 1;
      for   (int f0 = 0; f0 < fn0; f0++)
	for (int y0 = 0; y0 < yn0; y0++) { x0 = y0 - f0 + fn0 - 1;
	switch (result) {
	case (kOutput):
	  yy[y1][y0] += ff[f1][f0] * xx[x1][x0];
	  break;
	case (kInput ):
	  xx[x1][x0] += yy[y1][y0] * ff[f1][f0];
	  break;
	case (kFilter):
	  ff[f1][f0] += xx[x1][x0] * yy[y1][y0];
	  break;
	}
	}}
  }

  static void apply3D(Rsf3 x, Rsf3 f, Rsf3 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int yn2 = ((RsfSpace) y.getSpace()).getNSamples(2);
    int x0, x1, x2;
    float[][][] yy = y.d, ff = f.d, xx = x.d;
    for       (int f2 = 0; f2 < fn2; f2++)
      for     (int y2 = 0; y2 < yn2; y2++) { x2 = y2 - f2 + fn2 - 1;
      for     (int f1 = 0; f1 < fn1; f1++)
	for   (int y1 = 0; y1 < yn1; y1++) { x1 = y1 - f1 + fn1 - 1;
	for   (int f0 = 0; f0 < fn0; f0++)
	  for (int y0 = 0; y0 < yn0; y0++) { x0 = y0 - f0 + fn0 - 1;
	  switch (result) {
	  case (kOutput):
	    yy[y2][y1][y0] += ff[f2][f1][f0] * xx[x2][x1][x0];
	    break;
	  case (kInput ):
	    xx[x2][x1][x0] += yy[y2][y1][y0] * ff[f2][f1][f0];
	    break;
	  case (kFilter):
	    ff[f2][f1][f0] += xx[x2][x1][x0] * yy[y2][y1][y0];
	    break;
	  }
	  }}}
  }

  static void apply4D(Rsf4 x, Rsf4 f, Rsf4 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int yn2 = ((RsfSpace) y.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int yn3 = ((RsfSpace) y.getSpace()).getNSamples(3);
    int x0, x1, x2, x3;
    float[][][][] yy = y.d, ff = f.d, xx = x.d;
    for	  	(int f3 = 0; f3 < fn3; f3++)
      for 	(int y3 = 0; y3 < yn3; y3++) { x3 = y3 - f3 + fn3 - 1;
      for      	(int f2 = 0; f2 < fn2; f2++)
	for    	(int y2 = 0; y2 < yn2; y2++) { x2 = y2 - f2 + fn2 - 1;
	for   	(int f1 = 0; f1 < fn1; f1++)
	  for 	(int y1 = 0; y1 < yn1; y1++) { x1 = y1 - f1 + fn1 - 1;
	  for	(int f0 = 0; f0 < fn0; f0++)
	    for (int y0 = 0; y0 < yn0; y0++) { x0 = y0 - f0 + fn0 - 1;
	    switch (result) {
	    case (kOutput):
	      yy[y3][y2][y1][y0] +=
		ff[f3][f2][f1][f0] * xx[x3][x2][x1][x0];
	      break;
	    case (kInput ):
	      xx[x3][x2][x1][x0] +=
		yy[y3][y2][y1][y0] * ff[f3][f2][f1][f0];
	      break;
	    case (kFilter):
	      ff[f3][f2][f1][f0] +=
		xx[x3][x2][x1][x0] * yy[y3][y2][y1][y0];
	      break;
	    }
	    }}}}
  }

  static void apply5D(Rsf5 x, Rsf5 f, Rsf5 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int yn2 = ((RsfSpace) y.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int yn3 = ((RsfSpace) y.getSpace()).getNSamples(3);
    int fn4 = ((RsfSpace) f.getSpace()).getNSamples(4);
    int yn4 = ((RsfSpace) y.getSpace()).getNSamples(4);
    int x0, x1, x2, x3, x4;
    float[][][][][] yy = y.d, ff = f.d, xx = x.d;
    for   	  (int f4 = 0; f4 < fn4; f4++)
      for 	  (int y4 = 0; y4 < yn4; y4++) { x4 = y4 - f4 + fn4 - 1;
      for	  (int f3 = 0; f3 < fn3; f3++)
	for       (int y3 = 0; y3 < yn3; y3++) { x3 = y3 - f3 + fn3 - 1;
	for       (int f2 = 0; f2 < fn2; f2++)
	  for     (int y2 = 0; y2 < yn2; y2++) { x2 = y2 - f2 + fn2 - 1;
	  for     (int f1 = 0; f1 < fn1; f1++)
	    for   (int y1 = 0; y1 < yn1; y1++) { x1 = y1 - f1 + fn1 - 1;
	    for	  (int f0 = 0; f0 < fn0; f0++)
	      for (int y0 = 0; y0 < yn0; y0++) { x0 = y0 - f0 + fn0 - 1;
	      switch (result) {
	      case (kOutput):
		yy[y4][y3][y2][y1][y0] +=
		  ff[f4][f3][f2][f1][f0] * xx[x4][x3][x2][x1][x0];
		break;
	      case (kInput ):
		xx[x4][x3][x2][x1][x0] +=
		  yy[y4][y3][y2][y1][y0] * ff[f4][f3][f2][f1][f0];
		break;
	      case (kFilter):
		ff[f4][f3][f2][f1][f0] +=
		  xx[x4][x3][x2][x1][x0] * yy[y4][y3][y2][y1][y0];
		break;
	      }
	      }}}}}
  }

  static void apply6D(Rsf6 x, Rsf6 f, Rsf6 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int yn2 = ((RsfSpace) y.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int yn3 = ((RsfSpace) y.getSpace()).getNSamples(3);
    int fn4 = ((RsfSpace) f.getSpace()).getNSamples(4);
    int yn4 = ((RsfSpace) y.getSpace()).getNSamples(4);
    int fn5 = ((RsfSpace) f.getSpace()).getNSamples(5);
    int yn5 = ((RsfSpace) y.getSpace()).getNSamples(5);
    int x0, x1, x2, x3, x4, x5;
    float[][][][][][] yy = y.d, ff = f.d, xx = x.d;
    for       	    (int f5 = 0; f5 < fn5; f5++)
      for     	    (int y5 = 0; y5 < yn5; y5++) { x5 = y5 - f5 + fn5 - 1;
      for   	    (int f4 = 0; f4 < fn4; f4++)
	for 	    (int y4 = 0; y4 < yn4; y4++) { x4 = y4 - f4 + fn4 - 1;
	for	    (int f3 = 0; f3 < fn3; f3++)
	  for       (int y3 = 0; y3 < yn3; y3++) { x3 = y3 - f3 + fn3 - 1;
	  for       (int f2 = 0; f2 < fn2; f2++)
	    for     (int y2 = 0; y2 < yn2; y2++) { x2 = y2 - f2 + fn2 - 1;
	    for     (int f1 = 0; f1 < fn1; f1++)
	      for   (int y1 = 0; y1 < yn1; y1++) { x1 = y1 - f1 + fn1 - 1;
	      for   (int f0 = 0; f0 < fn0; f0++)
		for (int y0 = 0; y0 < yn0; y0++) { x0 = y0 - f0 + fn0 - 1;
		switch (result) {
		case (kOutput):
		  yy[y5][y4][y3][y2][y1][y0] +=
		    ff[f5][f4][f3][f2][f1][f0] * xx[x5][x4][x3][x2][x1][x0];
		  break;
		case (kInput ):
		  xx[x5][x4][x3][x2][x1][x0] +=
		    yy[y5][y4][y3][y2][y1][y0] * ff[f5][f4][f3][f2][f1][f0];
		  break;
		case (kFilter):
		  ff[f5][f4][f3][f2][f1][f0] +=
		    xx[x5][x4][x3][x2][x1][x0] * yy[y5][y4][y3][y2][y1][y0];
		  break;
		}
		}}}}}}
  }

  static RsfSpace createInputSpace (RsfSpace filtspc ,
				     RsfSpace oputspc) {
    int n = oputspc.getDimension();
    Axis[] inputAxes = new Axis[ n];
    for (int i = 0 ; i < n ; i++) {
      inputAxes[i] = new Axis(((RsfSpace) oputspc).getAxes()[i]);
      inputAxes[i].nsamples =   oputspc.getNSamples(i)
                              + filtspc.getNSamples(i) - 1;
    }
    return new RsfSpace(inputAxes);
  }

  static RsfSpace createFilterSpace(RsfSpace iputspc ,
				     RsfSpace oputspc) {
    int n = oputspc.getDimension();
    int[] filterNSamples = new int[ n];
    for (int i = 0 ; i < n ; i++)
      filterNSamples[i] = iputspc.getNSamples(i) - oputspc.getNSamples(i) + 1;
    return new RsfSpace(filterNSamples);
  }

  static RsfSpace createOutputSpace (RsfSpace iputspc ,
				      RsfSpace filtspc) {
    int n = iputspc.getDimension();
    Axis[] outputAxes = new Axis[ n];
    for (int i = 0 ; i < n ; i++) {
      outputAxes[i] = new Axis(((RsfSpace) iputspc).getAxes()[i]);
      outputAxes[i].nsamples =   iputspc.getNSamples(i)
                               - filtspc.getNSamples(i) + 1; 
    }
    return new RsfSpace(outputAxes);
  }
}

