package conv.conviF;

import jam.vector.*;
import jam.operator.*;
import jam.JamObject;
import rsf.vector.Axis;
import rsfF.vector.*;
import sep.operator.SepBiLinearOperator;

/**
 * Internal convolution (as a bilinear operator)
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * <BR><BR>The map is   A(i,f) |---> o
 * <BR> where: <UL>
 * <LI> A = bilinear internal convolution operator
 * <LI> i = input
 * <LI> f = filter
 * <LI> o = output
 * </UL>
 * <BR><BR>
 * 
 * @author    Joel Schroeder
 */
public class InternalConvolution extends SepBiLinearOperator {
  public static final int kInput  = SepBiLinearOperator.kInput1;
  public static final int kFilter = SepBiLinearOperator.kInput2;
  public static final int kOutput = SepBiLinearOperator.kOutput;

  public InternalConvolution(int spc1, RsfSpace space1,
			     int spc2, RsfSpace space2) {
    super(null, null, null);

    if ((spc1 != spc2)
	&& setSpace(spc1, space1)
	&& setSpace(spc2, space2)) {
      if      (getRange()   == null)
	setRange(   createOutputSpace((RsfSpace) getDomain1(),
				      (RsfSpace) getDomain2()));
      else if (getDomain1() == null)
	setDomain1( createInputSpace ((RsfSpace) getDomain2(),
				      (RsfSpace) getRange()));
      else
	setDomain2( createFilterSpace((RsfSpace) getDomain1(),
				      (RsfSpace) getRange()));
      return;
    }

    System.err.println("Error: integers to InternalConvolution" + 
		       " constructor must be two of:"           + "\n" +
		       " kInput, kFilter, kOutput." + "\n" ); 
  }

  /** copy constructor */
  protected InternalConvolution(InternalConvolution src) {
    super(src);
  }

  /** The domain and range spaces are shallow copies */
  public JamObject copy() { return new InternalConvolution(this); }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (!(obj2 instanceof InternalConvolution )) return false;
    return super.equals( (InternalConvolution) obj2);
  }

  public Operator getOp1(Vector x1) { return getIcaf((Rsf) x1); }
  public Operator getOp2(Vector x2) { return getIcai((Rsf) x2); }

  public Icaf     getIcaf(Rsf input ) {
    if (!(getDomain1().hasMember(input)))
      System.err.println("Error in InternalConvolution.getIcaf(): " +
			 "Vector is not in domain1\n");
    return new IcafFactory(input).getIcafFromOutputSpace((RsfSpace) range);
  }

  public Icai     getIcai(Rsf filter) {
    if (!(getDomain2().hasMember(filter)))
      System.err.println("Error in InternalConvolution.getIcai(): " +
			 "Vector is not in domain2\n");
    return new IcaiFactory(filter).getIcaiFromOutputSpace((RsfSpace) range);
  }

  /**
   * <STRONG>WARNING: not yet implemented</STRONG>
   */
  public hasAdjoint getPartialAdjoint () {
    System.err.println("Error: InternalConvolution" +
			".getPartialAdjoint() not yet implemented.\n" );
    return null; }

  // I/O
  public String toString() {
    return ("Internal Convolution:\n" + super.toString());
  }

  protected void apply(boolean add, Vector x, Vector f,
		       Vector y, int result) {
    staticApply(add, x, f, y, result);
  }

  static void staticApply(boolean add, Vector x, Vector f,
			  Vector y, int result) {
    int dim = ((RsfSpace) x.getSpace()).getDimension();

    switch (result) {
    case kInput:
      if (!add) x.zero();
      break;
    case kFilter:
      if (!add) f.zero();
      break;
    case kOutput:
      if (!add) y.zero();
      break;
    default:
      System.err.println("Error: integer argument \"result\" is invalid:\n" +
			 "       result = " + result);
      break;
    }

    switch (dim) {
    case   1:
      apply1D((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    case   2:
      apply2D((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    default:
      if (dim > 10)
	System.err.println("Error: Tcai only supported" +
			   " up to 10 dimensions");
      else
	applyManyD((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    }
  }

  static void apply1D(Rsf x, Rsf f, Rsf y, int result) {
    int fn = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn = ((RsfSpace) y.getSpace()).getNSamples(0);
    int x0, f0, y0;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for   ( f0 = 0; f0 < fn; f0++ )
      for ( y0 = 0; y0 < yn; y0++ ) {
	x0 = y0 - f0 + fn - 1;
	switch (result) {
	case (kOutput):
	  yy[y0] += ff[f0] * xx[x0];
	  break;
	case (kInput ):
	  xx[x0] += yy[y0] * ff[f0];
	  break;
	case (kFilter):
	  ff[f0] += xx[x0] * yy[y0];
	  break;
	}
      }
  }

  static void apply2D(Rsf x, Rsf f, Rsf y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int yn0 = ((RsfSpace) y.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int yn1 = ((RsfSpace) y.getSpace()).getNSamples(1);
    int x0, x1;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for       ( int f1 = 0; f1 < fn1; f1++ )
      for     ( int y1 = 0; y1 < yn1; y1++ ) { x1 = y1 - f1 + fn1 - 1;
	for   ( int f0 = 0; f0 < fn0; f0++ )
	  for ( int y0 = 0; y0 < yn0; y0++ ) { x0 = y0 - f0 + fn0 - 1;
	  switch (result) {
	  case (kOutput):
	    yy[y.i(y0,y1)] += ff[f.i(f0,f1)] * xx[x.i(x0,x1)];
	    break;
	  case (kInput ):
	    xx[x.i(x0,x1)] += yy[y.i(y0,y1)] * ff[f.i(f0,f1)];
	    break;
	  case (kFilter):
	    ff[f.i(f0,f1)] += xx[x.i(x0,x1)] * yy[y.i(y0,y1)];
	    break;
	  }
	  }}
  }

  static void applyManyD(Rsf x, Rsf f, Rsf y, int result) {
    int[] fn = new int[10], yn = new int[10];
    for ( int i = 0; i < 10; i++ ) {
      fn[i] = ((RsfSpace) f.getSpace()).getNSamples(i);
      yn[i] = ((RsfSpace) y.getSpace()).getNSamples(i);
    }
    float[] yy = y.d, ff = f.d, xx = x.d;
    int[] xi = new int[10], fi = new int[10], yi = new int[10];
    for (fi[9] = 0; fi[9] < fn[9]; fi[9]++)
    for (yi[9] = 0; yi[9] < yn[9]; yi[9]++) { xi[9] = yi[9] -fi[9] +fn[9] -1;
    for (fi[8] = 0; fi[8] < fn[8]; fi[8]++)		     	     	     
    for (yi[8] = 0; yi[8] < yn[8]; yi[8]++) { xi[8] = yi[8] -fi[8] +fn[8] -1;
    for (fi[7] = 0; fi[7] < fn[7]; fi[7]++)		     	     	     
    for (yi[7] = 0; yi[7] < yn[7]; yi[7]++) { xi[7] = yi[7] -fi[7] +fn[7] -1;
    for (fi[6] = 0; fi[6] < fn[6]; fi[6]++)		     	     	     
    for (yi[6] = 0; yi[6] < yn[6]; yi[6]++) { xi[6] = yi[6] -fi[6] +fn[6] -1;
    for (fi[5] = 0; fi[5] < fn[5]; fi[5]++)		     	     	     
    for (yi[5] = 0; yi[5] < yn[5]; yi[5]++) { xi[5] = yi[5] -fi[5] +fn[5] -1;
    for (fi[4] = 0; fi[4] < fn[4]; fi[4]++)		     	     	     
    for (yi[4] = 0; yi[4] < yn[4]; yi[4]++) { xi[4] = yi[4] -fi[4] +fn[4] -1;
    for (fi[3] = 0; fi[3] < fn[3]; fi[3]++)		     	     	     
    for (yi[3] = 0; yi[3] < yn[3]; yi[3]++) { xi[3] = yi[3] -fi[3] +fn[3] -1;
    for (fi[2] = 0; fi[2] < fn[2]; fi[2]++)		     	     	     
    for (yi[2] = 0; yi[2] < yn[2]; yi[2]++) { xi[2] = yi[2] -fi[2] +fn[2] -1;
    for (fi[1] = 0; fi[1] < fn[1]; fi[1]++)		     	     	     
    for (yi[1] = 0; yi[1] < yn[1]; yi[1]++) { xi[1] = yi[1] -fi[1] +fn[1] -1;
    for (fi[0] = 0; fi[0] < fn[0]; fi[0]++)		     	     	     
    for (yi[0] = 0; yi[0] < yn[0]; yi[0]++) { xi[0] = yi[0] -fi[0] +fn[0] -1;
    switch (result) {
    case (kOutput):
      yy[y.i(yi)] += ff[f.i(fi)] * xx[x.i(xi)];
      break;
    case (kInput ):
      xx[x.i(xi)] += yy[y.i(yi)] * ff[f.i(fi)];
      break;
    case (kFilter):
      ff[f.i(fi)] += xx[x.i(xi)] * yy[y.i(yi)];
      break;
    }
    }}}}}}}}}}
  }

  static RsfSpace createInputSpace ( RsfSpace filtspc ,
				     RsfSpace oputspc ) {
    int n = oputspc.getDimension();
    Axis[] inputAxes = new Axis[ n];
    for ( int i = 0 ; i < n ; i++ ) {
      inputAxes[i] = new Axis( ((RsfSpace) oputspc).getAxes()[i] );
      inputAxes[i].nsamples =   oputspc.getNSamples(i)
                              + filtspc.getNSamples(i) - 1;
    }
    return new RsfSpace( inputAxes ); }

  static RsfSpace createFilterSpace( RsfSpace iputspc ,
				     RsfSpace oputspc ) {
    int n = oputspc.getDimension();
    int[] filterNSamples = new int[ n];
    for ( int i = 0 ; i < n ; i++ )
      filterNSamples[i] = iputspc.getNSamples(i) - oputspc.getNSamples(i) + 1;
    return new RsfSpace( filterNSamples ); }

  static RsfSpace createOutputSpace ( RsfSpace iputspc ,
				      RsfSpace filtspc ) {
    int n = iputspc.getDimension();
    Axis[] outputAxes = new Axis[ n];
    for ( int i = 0 ; i < n ; i++ ) {
      outputAxes[i] = new Axis( ((RsfSpace) iputspc).getAxes()[i] );
      outputAxes[i].nsamples =   iputspc.getNSamples(i)
                               - filtspc.getNSamples(i) + 1; 
    }
    return new RsfSpace( outputAxes ); }

}

