package conv.convt;

import 	 jam.vector.*;
import 	 jam.operator.*;
import 	 jam.JamObject;
import  rsf.vector.*;
import   rsf.vector.Axis;
import   sep.operator.SepBiLinearOperator;

/**
 * Transient convolution (as a bilinear operator)
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * <BR><BR>The map is   A(i,f) |---> o
 * <BR> where: <UL>
 * <LI> A = bilinear transient convolution operator
 * <LI> i = input
 * <LI> f = filter
 * <LI> o = output
 * </UL>
 * 
 * @author    Joel Schroeder
 */
public class TransientConvolution extends SepBiLinearOperator {
  public static final int kInput  = SepBiLinearOperator.kInput1;
  public static final int kFilter = SepBiLinearOperator.kInput2;
  public static final int kOutput = SepBiLinearOperator.kOutput;
  
  public TransientConvolution(int spc1, RsfSpace space1,
			      int spc2, RsfSpace space2) {
    super(null, null, null);
    
    if ((spc1 != spc2)
	&& setSpace(spc1, space1)
	&& setSpace(spc2, space2)) {
      if      (getRange()   == null)
	setRange(  createOutputSpace((RsfSpace) getDomain1(),
				     (RsfSpace) getDomain2()));
      else if (getDomain1() == null)
	setDomain1(createInputSpace ((RsfSpace) getDomain2(),
				     (RsfSpace) getRange()));
      else
	setDomain2(createFilterSpace((RsfSpace) getDomain1(),
				     (RsfSpace) getRange()));
      return;
    }
    
    System.err.println("Error: integers to TransientConvolution" + 
		       " constructor must be two of:"           + "\n" +
		       " kInput, kFilter, kOutput." + "\n"); 
  }
  
  /** copy constructor */
  protected TransientConvolution(TransientConvolution src) {
    super(src);
  }
  
  /** The domain and range spaces are shallow copies */
  public JamObject copy() { return new TransientConvolution(this); }
  
  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (!(obj2 instanceof TransientConvolution)) return false;
    return super.equals(obj2);
  }
  
  public Operator getOp1(Vector x1) { return getTcaf((Rsf) x1); }
  public Operator getOp2(Vector x2) { return getTcai((Rsf) x2); }
  
  public Tcaf                  getTcaf(Rsf input ) {
    //    if (!(domain1.hasMember(input)))
    //      System.err.println("Error: Vector is not in domain1\n");
    return new TcafFactory(input).getTcafFromOutputSpace((RsfSpace) range);
  }
  
  public Tcai                  getTcai(Rsf filter) {
    //    if (!(domain2.hasMember(filter)))
    //      System.err.println("Error: Vector is not in domain2\n");
    return new TcaiFactory(filter).getTcaiFromOutputSpace((RsfSpace) range);
  }
  
  /**
   * <STRONG>WARNING: not yet implemented</STRONG>
   */
  public hasAdjoint getPartialAdjoint () {
    System.err.println("Error: TransientConvolution" +
		       ".getPartialAdjoint() not yet implemented.\n");
    return null; }
  
  // I/O
  public String toString() {
    return ("Transient Convolution:\n" + super.toString());
  }
  
  protected void apply(boolean add, Vector x, Vector f,
		       Vector y, int result) {
    staticApply(add, x, f, y, result);
  }
  
  static void staticApply(boolean add, Vector x, Vector f,
			  Vector y, int result) {
    int dim = ((RsfSpace) x.getSpace()).getDimension();
    
    switch (result) {
    case kInput:
      if (!add) x.zero();
      break;
    case kFilter:
      if (!add) f.zero();
      break;
    case kOutput:
      if (!add) y.zero();
      break;
    default:
      System.err.println("Error: integer argument \"result\" is invalid:\n" +
			 "       result = " + result);
      break;
    }

    switch (dim) {
    case   1:
      apply1D((Rsf1) x, (Rsf1) f, (Rsf1) y, result);
      break;
    case   2:
      apply2D((Rsf2) x, (Rsf2) f, (Rsf2) y, result);
      break;
    case 3:
      apply3D((Rsf3) x, (Rsf3) f, (Rsf3) y, result);
      break;
    case 4:
      apply4D((Rsf4) x, (Rsf4) f, (Rsf4) y, result);
      break;
    case 5:
      apply5D((Rsf5) x, (Rsf5) f, (Rsf5) y, result);
      break;
    case 6:
      apply6D((Rsf6) x, (Rsf6) f, (Rsf6) y, result);
      break;
    default:
      System.err.println("Error: Transient Convolution only supported" +
		         " up to 6 dimensions");
      break;
    }
  }
  
  static void apply1D(Rsf1 x, Rsf1 f, Rsf1 y, int result) {
    int fn = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn = ((RsfSpace) x.getSpace()).getNSamples(0);
    int y0;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for   (int f0 = 0; f0 < fn; f0++)
      for (int x0 = 0; x0 < xn; x0++) {	y0 = x0 + f0;
      switch (result) {
      case (kOutput):
	yy[y0] += ff[f0] * xx[x0];
	break;
      case (kInput ):
	xx[x0] += yy[y0] * ff[f0];
	break;
      case (kFilter):
	ff[f0] += xx[x0] * yy[y0];
	break;
      }
      }
  }
  
  static void apply2D(Rsf2 x, Rsf2 f, Rsf2 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int y0, y1;
    float[][] yy = y.d, ff = f.d, xx = x.d;
    for     (int f1 = 0; f1 < fn1; f1++)
      for   (int x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
      for   (int f0 = 0; f0 < fn0; f0++)
	for (int x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
	switch (result) {
	case (kOutput):
	  yy[y1][y0] += ff[f1][f0] * xx[x1][x0];
	  break;
	case (kInput ):
	  xx[x1][x0] += yy[y1][y0] * ff[f1][f0];
	  break;
	case (kFilter):
	  ff[f1][f0] += xx[x1][x0] * yy[y1][y0];
	  break;
	}
	}}
  }

  static void apply3D(Rsf3 x, Rsf3 f, Rsf3 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int xn2 = ((RsfSpace) x.getSpace()).getNSamples(2);
    int y0, y1, y2;
    float[][][] yy = y.d, ff = f.d, xx = x.d;
    for       (int f2 = 0; f2 < fn2; f2++)	       
      for     (int x2 = 0; x2 < xn2; x2++) { y2 = x2 + f2;
      for     (int f1 = 0; f1 < fn1; f1++)	       
	for   (int x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
	for   (int f0 = 0; f0 < fn0; f0++)	       
	  for (int x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
	  switch (result) {
	  case (kOutput):
	    yy[y2][y1][y0] += ff[f2][f1][f0] * xx[x2][x1][x0];
	    break;
	  case (kInput ):
	    xx[x2][x1][x0] += yy[y2][y1][y0] * ff[f2][f1][f0];
	    break;
	  case (kFilter):
	    ff[f2][f1][f0] += xx[x2][x1][x0] * yy[y2][y1][y0];
	    break;
	  }
	  }}}
  }
  
  static void apply4D(Rsf4 x, Rsf4 f, Rsf4 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int xn2 = ((RsfSpace) x.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int xn3 = ((RsfSpace) x.getSpace()).getNSamples(3);
    int y0, y1, y2, y3;
    float[][][][] yy = y.d, ff = f.d, xx = x.d;
    for	        (int f3 = 0; f3 < fn3; f3++)	       
      for       (int x3 = 0; x3 < xn3; x3++) { y3 = x3 + f3;
      for       (int f2 = 0; f2 < fn2; f2++)	       
	for     (int x2 = 0; x2 < xn2; x2++) { y2 = x2 + f2;
	for     (int f1 = 0; f1 < fn1; f1++)	       
	  for   (int x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
	  for   (int f0 = 0; f0 < fn0; f0++)	       
	    for (int x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
	    switch (result) {
	    case (kOutput):
	      yy[y3][y2][y1][y0] += ff[f3][f2][f1][f0] * xx[x3][x2][x1][x0];
	      break;
	    case (kInput ):
	      xx[x3][x2][x1][x0] += yy[y3][y2][y1][y0] * ff[f3][f2][f1][f0];
	      break;
	    case (kFilter):
	      ff[f3][f2][f1][f0] += xx[x3][x2][x1][x0] * yy[y3][y2][y1][y0];
	      break;
	    }
	    }}}}
  }
  
  static void apply5D(Rsf5 x, Rsf5 f, Rsf5 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int xn2 = ((RsfSpace) x.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int xn3 = ((RsfSpace) x.getSpace()).getNSamples(3);
    int fn4 = ((RsfSpace) f.getSpace()).getNSamples(4);
    int xn4 = ((RsfSpace) x.getSpace()).getNSamples(4);
    int y0, y1, y2, y3, y4;
    float[][][][][] yy = y.d, ff = f.d, xx = x.d;
    for   	  (int f4 = 0; f4 < fn4; f4++)	       
      for 	  (int x4 = 0; x4 < xn4; x4++) { y4 = x4 + f4;
      for	  (int f3 = 0; f3 < fn3; f3++)	       
	for       (int x3 = 0; x3 < xn3; x3++) { y3 = x3 + f3;
	for       (int f2 = 0; f2 < fn2; f2++)	       
	  for     (int x2 = 0; x2 < xn2; x2++) { y2 = x2 + f2;
	  for     (int f1 = 0; f1 < fn1; f1++)	       
	    for   (int x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
	    for   (int f0 = 0; f0 < fn0; f0++)	       
	      for (int x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
	      switch (result) {
	      case (kOutput):
		yy[y4][y3][y2][y1][y0] +=
		  ff[f4][f3][f2][f1][f0] * xx[x4][x3][x2][x1][x0];
		break;
	      case (kInput ):
		xx[x4][x3][x2][x1][x0] +=
		  yy[y4][y3][y2][y1][y0] * ff[f4][f3][f2][f1][f0];
		break;
	      case (kFilter):
		ff[f4][f3][f2][f1][f0] +=
		  xx[x4][x3][x2][x1][x0] * yy[y4][y3][y2][y1][y0];
		break;
	      }
	      }}}}}
  }
  
  static void apply6D(Rsf6 x, Rsf6 f, Rsf6 y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int fn2 = ((RsfSpace) f.getSpace()).getNSamples(2);
    int xn2 = ((RsfSpace) x.getSpace()).getNSamples(2);
    int fn3 = ((RsfSpace) f.getSpace()).getNSamples(3);
    int xn3 = ((RsfSpace) x.getSpace()).getNSamples(3);
    int fn4 = ((RsfSpace) f.getSpace()).getNSamples(4);
    int xn4 = ((RsfSpace) x.getSpace()).getNSamples(4);
    int fn5 = ((RsfSpace) f.getSpace()).getNSamples(5);
    int xn5 = ((RsfSpace) x.getSpace()).getNSamples(5);
    int y0, y1, y2, y3, y4, y5;
    float[][][][][][] yy = y.d, ff = f.d, xx = x.d;
    for       	    (int f5 = 0; f5 < fn5; f5++)
      for     	    (int x5 = 0; x5 < xn5; x5++) { y5 = x5 + f5;
      for   	    (int f4 = 0; f4 < fn4; f4++)	       
	for 	    (int x4 = 0; x4 < xn4; x4++) { y4 = x4 + f4;
	for	    (int f3 = 0; f3 < fn3; f3++)	       
	  for       (int x3 = 0; x3 < xn3; x3++) { y3 = x3 + f3;
	  for       (int f2 = 0; f2 < fn2; f2++)	       
	    for     (int x2 = 0; x2 < xn2; x2++) { y2 = x2 + f2;
	    for     (int f1 = 0; f1 < fn1; f1++)	       
	      for   (int x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
	      for   (int f0 = 0; f0 < fn0; f0++)	       
		for (int x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
		switch (result) {
		case (kOutput):
		  yy[y5][y4][y3][y2][y1][y0] +=
		    ff[f5][f4][f3][f2][f1][f0] * xx[x5][x4][x3][x2][x1][x0];
		  break;
		case (kInput ):
		  xx[x5][x4][x3][x2][x1][x0] +=
		    yy[y5][y4][y3][y2][y1][y0] * ff[f5][f4][f3][f2][f1][f0];
		  break;
		case (kFilter):
		  ff[f5][f4][f3][f2][f1][f0] +=
		    xx[x5][x4][x3][x2][x1][x0] * yy[y5][y4][y3][y2][y1][y0];
		  break;
		}
		}}}}}}
  }
  
  static RsfSpace createInputSpace (RsfSpace filtspc ,
				    RsfSpace oputspc) {
    int n = oputspc.getDimension();
    Axis[] inputAxes = new Axis[n];
    for (int i = 0 ; i < n ; i++) {
      inputAxes[i] = new Axis(((RsfSpace) oputspc).getAxes()[i]);
      inputAxes[i].nsamples =
	oputspc.getNSamples(i) -
	filtspc.getNSamples(i) + 1;
    }
    return new RsfSpace(inputAxes);
  }
  
  static RsfSpace createFilterSpace(RsfSpace iputspc ,
				    RsfSpace oputspc) {
    int n = oputspc.getDimension();
    int[] filterNSamples = new int[n];
    for (int i = 0 ; i < n ; i++)
      filterNSamples[i] =
	oputspc.getNSamples(i) -
	iputspc.getNSamples(i) + 1;
    return new RsfSpace(filterNSamples);
  }
  
  static RsfSpace createOutputSpace (RsfSpace iputspc ,
				     RsfSpace filtspc ) {
    int n = iputspc.getDimension();
    Axis[] outputAxes = new Axis[n];
    for (int i = 0 ; i < n ; i++) {
      outputAxes[i] = new Axis(((RsfSpace) iputspc).getAxes()[i]);
      outputAxes[i].nsamples =
	iputspc.getNSamples(i) +
	filtspc.getNSamples(i) - 1; 
    }
    return new RsfSpace(outputAxes);
  }
}

