package conv.convtF;

import jam.vector.*;
import jam.operator.*;
import jam.JamObject;
import rsf.vector.Axis;
import rsfF.vector.*;
import sep.operator.SepBiLinearOperator;

/**
 * Transient convolution (as a bilinear operator)
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * <BR><BR>The map is   A(i,f) |---> o
 * <BR> where: <UL>
 * <LI> A = bilinear transient convolution operator
 * <LI> i = input
 * <LI> f = filter
 * <LI> o = output
 * </UL>
 * 
 * @author    Joel Schroeder
 */
public class TransientConvolution extends SepBiLinearOperator {
  public static final int kInput  = SepBiLinearOperator.kInput1;
  public static final int kFilter = SepBiLinearOperator.kInput2;
  public static final int kOutput = SepBiLinearOperator.kOutput;

  public TransientConvolution(int spc1, RsfSpace space1,
			      int spc2, RsfSpace space2) {
    super(null, null, null);

    if ((spc1 != spc2)
	&& setSpace(spc1, space1)
	&& setSpace(spc2, space2)) {
      if      (getRange()   == null)
	setRange(   createOutputSpace((RsfSpace) getDomain1(),
				      (RsfSpace) getDomain2()));
      else if (getDomain1() == null)
	setDomain1( createInputSpace ((RsfSpace) getDomain2(),
				      (RsfSpace) getRange()));
      else
	setDomain2( createFilterSpace((RsfSpace) getDomain1(),
				      (RsfSpace) getRange()));
      return;
    }

    System.err.println("Error: integers to TransientConvolution" + 
		       " constructor must be two of:"           + "\n" +
		       " kInput, kFilter, kOutput." + "\n" ); 
  }

  /** copy constructor */
  protected TransientConvolution(TransientConvolution src) {
    super(src);
  }

  /** The domain and range spaces are shallow copies */
  public JamObject copy() { return new TransientConvolution(this); }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (!(obj2 instanceof TransientConvolution )) return false;
    return super.equals(obj2);
  }

  public Operator getOp1(Vector x1) { return getTcaf((Rsf) x1); }
  public Operator getOp2(Vector x2) { return getTcai((Rsf) x2); }

  public Tcaf                  getTcaf(Rsf input ) {
    //    if (!(domain1.hasMember(input)))
    //      System.err.println("Error: Vector is not in domain1\n");
    return new TcafFactory(input).getTcafFromOutputSpace((RsfSpace) range);
  }

  public Tcai                  getTcai(Rsf filter) {
    //    if (!(domain2.hasMember(filter)))
    //      System.err.println("Error: Vector is not in domain2\n");
    return new TcaiFactory(filter).getTcaiFromOutputSpace((RsfSpace) range);
  }

  /**
   * <STRONG>WARNING: not yet implemented</STRONG>
   */
  public hasAdjoint getPartialAdjoint () {
    System.err.println("Error: TransientConvolution" +
			".getPartialAdjoint() not yet implemented.\n" );
    return null; }

  // I/O
  public String toString() {
    return ("Transient Convolution:\n" + super.toString());
  }

  protected void apply(boolean add, Vector x, Vector f,
		       Vector y, int result) {
    staticApply(add, x, f, y, result);
  }

  static void staticApply(boolean add, Vector x, Vector f,
			  Vector y, int result) {
    int dim = ((RsfSpace) x.getSpace()).getDimension();

    switch (result) {
    case kInput:
      if (!add) x.zero();
      break;
    case kFilter:
      if (!add) f.zero();
      break;
    case kOutput:
      if (!add) y.zero();
      break;
    default:
      System.err.println("Error: integer argument \"result\" is invalid:\n" +
			 "       result = " + result);
      break;
    }

    switch (dim) {
    case   1:
      apply1D((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    case   2:
      apply2D((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    default:
      if (dim > 10)
	System.err.println("Error: Tcai only supported" +
			   " up to 10 dimensions");
      else
	applyManyD((Rsf) x, (Rsf) f, (Rsf) y, result);
      break;
    }
  }

  static void apply1D(Rsf x, Rsf f, Rsf y, int result) {
    int fn = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn = ((RsfSpace) x.getSpace()).getNSamples(0);
    int x0, f0, y0;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for   ( f0 = 0; f0 < fn; f0++ )
      for ( x0 = 0; x0 < xn; x0++ ) {
	y0 = x0 + f0;
	switch (result) {
	case (kOutput):
	  yy[y0] += ff[f0] * xx[x0];
	  break;
	case (kInput ):
	  xx[x0] += yy[y0] * ff[f0];
	  break;
	case (kFilter):
	  ff[f0] += xx[x0] * yy[y0];
	  break;
	}
      }
  }

  static void apply2D(Rsf x, Rsf f, Rsf y, int result) {
    int fn0 = ((RsfSpace) f.getSpace()).getNSamples(0);
    int xn0 = ((RsfSpace) x.getSpace()).getNSamples(0);
    int fn1 = ((RsfSpace) f.getSpace()).getNSamples(1);
    int xn1 = ((RsfSpace) x.getSpace()).getNSamples(1);
    int x0, x1, f0, f1, y0, y1;
    float[] yy = y.d, ff = f.d, xx = x.d;
    for       (f1 = 0; f1 < fn1; f1++)
      for     (x1 = 0; x1 < xn1; x1++) { y1 = x1 + f1;
	for   (f0 = 0; f0 < fn0; f0++)
	  for (x0 = 0; x0 < xn0; x0++) { y0 = x0 + f0;
	  switch (result) {
	  case (kOutput):
	    yy[y.i(y0,y1)] += ff[f.i(f0,f1)] * xx[x.i(x0,x1)];
	    break;
	  case (kInput ):
	    xx[x.i(x0,x1)] += yy[y.i(y0,y1)] * ff[f.i(f0,f1)];
	    break;
	  case (kFilter):
	    ff[f.i(f0,f1)] += xx[x.i(x0,x1)] * yy[y.i(y0,y1)];
	    break;
	  }
	  }}
  }

  static void applyManyD(Rsf x, Rsf f, Rsf y, int result) {
    int[] fn = new int[10], xn = new int[10];
    for ( int i = 0; i < 10; i++ ) {
      fn[i] = ((RsfSpace) f.getSpace()).getNSamples(i);
      xn[i] = ((RsfSpace) x.getSpace()).getNSamples(i);
    }
    float[] yy = y.d, ff = f.d, xx = x.d;
    int[] xi = new int[10], fi = new int[10], yi = new int[10];
    for (fi[9] = 0; fi[9] < fn[9]; fi[9]++)
    for (xi[9] = 0; xi[9] < xn[9]; xi[9]++) { yi[9] = xi[9] + fi[9];
    for (fi[8] = 0; fi[8] < fn[8]; fi[8]++)		     	     	     
    for (xi[8] = 0; xi[8] < xn[8]; xi[8]++) { yi[8] = xi[8] + fi[8];
    for (fi[7] = 0; fi[7] < fn[7]; fi[7]++)		     	     	     
    for (xi[7] = 0; xi[7] < xn[7]; xi[7]++) { yi[7] = xi[7] + fi[7];
    for (fi[6] = 0; fi[6] < fn[6]; fi[6]++)		     	     	     
    for (xi[6] = 0; xi[6] < xn[6]; xi[6]++) { yi[6] = xi[6] + fi[6];
    for (fi[5] = 0; fi[5] < fn[5]; fi[5]++)		     	     	     
    for (xi[5] = 0; xi[5] < xn[5]; xi[5]++) { yi[5] = xi[5] + fi[5];
    for (fi[4] = 0; fi[4] < fn[4]; fi[4]++)		     	     	     
    for (xi[4] = 0; xi[4] < xn[4]; xi[4]++) { yi[4] = xi[4] + fi[4];
    for (fi[3] = 0; fi[3] < fn[3]; fi[3]++)		     	     	     
    for (xi[3] = 0; xi[3] < xn[3]; xi[3]++) { yi[3] = xi[3] + fi[3];
    for (fi[2] = 0; fi[2] < fn[2]; fi[2]++)		     	     	     
    for (xi[2] = 0; xi[2] < xn[2]; xi[2]++) { yi[2] = xi[2] + fi[2];
    for (fi[1] = 0; fi[1] < fn[1]; fi[1]++)		     	     	     
    for (xi[1] = 0; xi[1] < xn[1]; xi[1]++) { yi[1] = xi[1] + fi[1];
    for (fi[0] = 0; fi[0] < fn[0]; fi[0]++)		     	     	     
    for (xi[0] = 0; xi[0] < xn[0]; xi[0]++) { yi[0] = xi[0] + fi[0];
    switch (result) {
    case (kOutput):
      yy[y.i(yi)] += ff[f.i(fi)] * xx[x.i(xi)];
      break;
    case (kInput ):
      xx[x.i(xi)] += yy[y.i(yi)] * ff[f.i(fi)];
      break;
    case (kFilter):
      ff[f.i(fi)] += xx[x.i(xi)] * yy[y.i(yi)];
      break;
    }
    }}}}}}}}}}
  }

  static RsfSpace createInputSpace ( RsfSpace filtspc ,
				     RsfSpace oputspc ) {
    int n = oputspc.getDimension();
    Axis[] inputAxes = new Axis[ n];
    for ( int i = 0 ; i < n ; i++ ) {
      inputAxes[i] = new Axis( ((RsfSpace) oputspc).getAxes()[i] );
      inputAxes[i].nsamples =
	oputspc.getNSamples(i) -
	filtspc.getNSamples(i) + 1;
    }
    return new RsfSpace( inputAxes );
  }

  static RsfSpace createFilterSpace( RsfSpace iputspc ,
				     RsfSpace oputspc ) {
    int n = oputspc.getDimension();
    int[] filterNSamples = new int[ n];
    for ( int i = 0 ; i < n ; i++ )
      filterNSamples[i] =
	oputspc.getNSamples(i) -
	iputspc.getNSamples(i) + 1;
    return new RsfSpace( filterNSamples );
  }

  static RsfSpace createOutputSpace ( RsfSpace iputspc ,
				      RsfSpace filtspc ) {
    int n = iputspc.getDimension();
    Axis[] outputAxes = new Axis[ n];
    for ( int i = 0 ; i < n ; i++ ) {
      outputAxes[i] = new Axis( ((RsfSpace) iputspc).getAxes()[i] );
      outputAxes[i].nsamples =
	iputspc.getNSamples(i) +
	filtspc.getNSamples(i) - 1; 
    }
    return new RsfSpace( outputAxes );
  }
}

